#include "stdafx.h"
#include "ThrowOnFail.h"

LRESULT CALLBACK WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

int WINAPI wWinMain(HINSTANCE hInstance, HINSTANCE, PWSTR pCmdLine, int nCmdShow)
{
	// Register the window class.
	const wchar_t CLASS_NAME[] = L"D3D11 Window";

	WNDCLASS wc = {};

	wc.lpfnWndProc = WindowProc;
	wc.hInstance = hInstance;
	wc.lpszClassName = CLASS_NAME;

	RegisterClass(&wc);

	// Create the window.

	HWND hwnd = CreateWindowEx(
		0,                              // Optional window styles.
		CLASS_NAME,                     // Window class
		L"GraphGame17",					// Window text
		WS_OVERLAPPEDWINDOW,            // Window style
										// Size and position
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		NULL,       // Parent window    
		NULL,       // Menu
		hInstance,  // Instance handle
		NULL        // Additional application data
		);

	if (hwnd == NULL)
	{
		return 0;
	}

	ShowWindow(hwnd, nCmdShow);

	UINT creationFlags = D3D11_CREATE_DEVICE_BGRA_SUPPORT;
#if defined(_DEBUG)
	creationFlags |= D3D11_CREATE_DEVICE_DEBUG;
#endif

	D3D_FEATURE_LEVEL featureLevels[] =
	{
		D3D_FEATURE_LEVEL_11_1,
		D3D_FEATURE_LEVEL_11_0
	};

	using namespace Microsoft::WRL;
	// Create the Direct3D 11 API device object and a corresponding context.
	ComPtr<ID3D11Device> device;
	ComPtr<ID3D11DeviceContext> context;
	D3D_FEATURE_LEVEL d3dFeatureLevel;

	GraphGame::ThrowOnFail("Failed to create D3D device.", __FILE__, __LINE__) ^
		D3D11CreateDevice(
			nullptr,					// Specify nullptr to use the default adapter.
			D3D_DRIVER_TYPE_HARDWARE,	// Create a device using the hardware graphics driver.
			0,							// Should be 0 unless the driver is D3D_DRIVER_TYPE_SOFTWARE.
			creationFlags,				// Set debug and Direct2D compatibility flags.
			featureLevels,				// List of feature levels this app can support.
			ARRAYSIZE(featureLevels),	// Size of the list above.
			D3D11_SDK_VERSION,			// Always set this to D3D11_SDK_VERSION for Windows Store apps.
			device.GetAddressOf(),		// Returns the Direct3D device created.
			&d3dFeatureLevel,			// Returns feature level of device created.
			context.GetAddressOf()		// Returns the device immediate context.
			);

	ComPtr<IDXGIDevice3> dxgiDevice;
	GraphGame::ThrowOnFail("The device is not a DXGI device.", __FILE__, __LINE__) ^
		device.As(&dxgiDevice);

	ComPtr<IDXGIAdapter> dxgiAdapter;
	GraphGame::ThrowOnFail("Failed to get adapter from device.", __FILE__, __LINE__) ^
		dxgiDevice->GetAdapter(dxgiAdapter.GetAddressOf());

	ComPtr<IDXGIFactory2> dxgiFactory;
	GraphGame::ThrowOnFail("Failed to get factory from adapter.", __FILE__, __LINE__) ^
		dxgiAdapter->GetParent(IID_PPV_ARGS(dxgiFactory.GetAddressOf()));

	ComPtr<IDXGISwapChain1> swapChain;
	RECT r;
	GetWindowRect(hwnd, &r);

	DXGI_SWAP_CHAIN_DESC1 swapChainDesc = { 0 };
	swapChainDesc.Width = r.right - r.left; // Match the size of the window.
	swapChainDesc.Height = r.bottom - r.top;
	swapChainDesc.Format = DXGI_FORMAT_B8G8R8A8_UNORM; // This is the most common swap chain format.
	swapChainDesc.Stereo = false;
	swapChainDesc.SampleDesc.Count = 1; // Don't use multi-sampling.
	swapChainDesc.SampleDesc.Quality = 0;
	swapChainDesc.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;
	swapChainDesc.BufferCount = 2; // Use double-buffering to minimize latency.
	swapChainDesc.SwapEffect = DXGI_SWAP_EFFECT_FLIP_SEQUENTIAL; // All Windows Store apps must use this SwapEffect.
	swapChainDesc.Flags = 0;
	swapChainDesc.Scaling = DXGI_SCALING_NONE;
	swapChainDesc.AlphaMode = DXGI_ALPHA_MODE_IGNORE;

	GraphGame::ThrowOnFail("Failed to create swap chain.", __FILE__, __LINE__) ^
		dxgiFactory->CreateSwapChainForHwnd(
			device.Get(),
			hwnd,
			&swapChainDesc,
			nullptr,
			nullptr,
			&swapChain
			);

	ComPtr<ID3D11Buffer> vertexBuffer;
	ComPtr<ID3D11InputLayout> inputLayout;
	ComPtr<ID3D11VertexShader> vertexShader;
	ComPtr<ID3D11PixelShader> pixelShader;

	struct float3 {
		float x, y, z;
		float3(float x, float y, float z) :x(x), y(y), z(z) {}
	};

	D3D11_BUFFER_DESC desc;
	desc.BindFlags = D3D11_BIND_VERTEX_BUFFER;
	desc.ByteWidth = sizeof(float3) * 3;
	desc.CPUAccessFlags = 0;
	desc.MiscFlags = 0;
	desc.StructureByteStride = sizeof(float3);
	desc.Usage = D3D11_USAGE_IMMUTABLE;

	float3 vertexPositionArray[3] = {
		float3(0, 0, 0.5),
		float3(0, 1, 0.5),
		float3(1, 0, 0.5) };
	D3D11_SUBRESOURCE_DATA initData;
	initData.pSysMem = vertexPositionArray;
	initData.SysMemPitch = 0;
	initData.SysMemSlicePitch = 0;

	GraphGame::ThrowOnFail("Failed to create vertex buffer.", __FILE__, __LINE__) ^
		device->CreateBuffer(&desc, &initData, vertexBuffer.GetAddressOf());

	const char* vertexShaderCode =
		"float4 vsIdle(float4 pos :POSITION ) :SV_Position {return pos;}";
	ID3DBlob* vertexShaderByteCode;
	GraphGame::ThrowOnFail("Failed to compile vertex shader source.", __FILE__, __LINE__) ^
		D3DCompile(vertexShaderCode, strlen(vertexShaderCode), NULL, NULL, NULL, "vsIdle", "vs_5_0", 0, 0, &vertexShaderByteCode, NULL);
	
	GraphGame::ThrowOnFail("Failed to create vertex shader.", __FILE__, __LINE__) ^
		device->CreateVertexShader(
			vertexShaderByteCode->GetBufferPointer(), vertexShaderByteCode->GetBufferSize(), NULL, vertexShader.GetAddressOf());

	// insert after #2.7
	D3D11_INPUT_ELEMENT_DESC positionElement;
	positionElement.AlignedByteOffset = 0;
	positionElement.Format = DXGI_FORMAT_R32G32B32_FLOAT;
	positionElement.InputSlot = 0;
	positionElement.InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
	positionElement.InstanceDataStepRate = 0;
	positionElement.SemanticIndex = 0;
	positionElement.SemanticName = "POSITION";

	GraphGame::ThrowOnFail("Failed to create input layout.", __FILE__, __LINE__) ^
		device->CreateInputLayout(
			&positionElement, 1, vertexShaderByteCode
			->GetBufferPointer(), vertexShaderByteCode
			->GetBufferSize(), inputLayout.GetAddressOf());

	const char* pixelShaderCode = "float4 psIdle() : SV_Target {return float4(1, 0, 0, 1);}";
	ID3DBlob* pixelShaderByteCode;
	GraphGame::ThrowOnFail("Failed to compile pixel shader code.", __FILE__, __LINE__) ^
		D3DCompile(pixelShaderCode, strlen(pixelShaderCode), NULL, NULL, NULL, "psIdle", "ps_5_0", 0, 0, &pixelShaderByteCode, NULL);
	GraphGame::ThrowOnFail("Failed to create pixel shader.", __FILE__, __LINE__) ^
		device->CreatePixelShader(
			pixelShaderByteCode->GetBufferPointer(), pixelShaderByteCode->GetBufferSize(), NULL, pixelShader.GetAddressOf());

	ComPtr<ID3D11Texture2D> backBuffer;
	GraphGame::ThrowOnFail("Failed to get swapchain backbuffer.", __FILE__, __LINE__) ^
		swapChain->GetBuffer(0, IID_PPV_ARGS(backBuffer.GetAddressOf()));

	ComPtr<ID3D11RenderTargetView> renderTargetView;
	GraphGame::ThrowOnFail("Failed to create render target view.", __FILE__, __LINE__) ^
		device->CreateRenderTargetView(
			backBuffer.Get(),
			nullptr,
			renderTargetView.GetAddressOf()
			);

	// Create a depth stencil view for use with 3D rendering if needed.
	CD3D11_TEXTURE2D_DESC depthStencilDesc(
		DXGI_FORMAT_D24_UNORM_S8_UINT,
		static_cast<UINT>(swapChainDesc.Width),
		static_cast<UINT>(swapChainDesc.Height),
		1, // This depth stencil view has only one texture.
		1, // Use a single mipmap level.
		D3D11_BIND_DEPTH_STENCIL
		);

	ComPtr<ID3D11Texture2D> depthStencil;
	GraphGame::ThrowOnFail("Failed to create depth stencil texture.", __FILE__, __LINE__) ^
		device->CreateTexture2D(
			&depthStencilDesc,
			nullptr,
			&depthStencil
		);

	ComPtr<ID3D11DepthStencilView> depthStencilView;
	CD3D11_DEPTH_STENCIL_VIEW_DESC depthStencilViewDesc(D3D11_DSV_DIMENSION_TEXTURE2D);
	GraphGame::ThrowOnFail("Failed to create depth stencil view.", __FILE__, __LINE__) ^
		device->CreateDepthStencilView(
			depthStencil.Get(),
			&depthStencilViewDesc,
			depthStencilView.GetAddressOf()
			);

	// Set the 3D rendering viewport to target the entire window.
	CD3D11_VIEWPORT screenViewport(
		0.0f,
		0.0f,
		swapChainDesc.Width,
		swapChainDesc.Height
		);

	context->RSSetViewports(1, &screenViewport);

	MSG msg = { 0 };

	while (WM_QUIT != msg.message)
	{
		if (PeekMessage(&msg, nullptr, 0, 0, PM_REMOVE))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		else
		{
			context->OMSetRenderTargets(1, renderTargetView.GetAddressOf(), depthStencilView.Get());
			// idle: update, render
			float clearColor[4] =
			{ 0.9f, 0.7f, 0.1f, 0.0f };
			context->ClearRenderTargetView(
				renderTargetView.Get(), clearColor);
			context->ClearDepthStencilView(depthStencilView.Get(), D3D11_CLEAR_DEPTH, 1.0, 0);

			unsigned int stride = sizeof(float3);
			unsigned int offset = 0;
			context->IASetVertexBuffers(0, 1, vertexBuffer.GetAddressOf(), &stride, &offset);
			context->IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST);
			context->IASetInputLayout(inputLayout.Get());
			context->VSSetShader(vertexShader.Get(),
				NULL, 0);
			context->PSSetShader(pixelShader.Get(), NULL, 0);

			context->Draw(3, 0);

			GraphGame::ThrowOnFail("Failed to present swap chain.", __FILE__, __LINE__) ^
				swapChain->Present(0, 0);
		}
	}

	return 0;
}

LRESULT CALLBACK WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	}
	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}
